@@ -5,9 +5,9 @@
 ;; Author: Omar Antolín Camarena <omar@matem.unam.mx>
 ;; Maintainer: Omar Antolín Camarena <omar@matem.unam.mx>, Daniel Mendler <mail@daniel-mendler.de>
 ;; Keywords: extensions
-;; Version: 1.1
+;; Version: 1.2
 ;; Homepage: https://github.com/oantolin/orderless
-;; Package-Requires: ((emacs "27.1"))
+;; Package-Requires: ((emacs "27.1") (compat "30"))
 
 ;; This file is part of GNU Emacs.
 
@@ -55,6 +55,7 @@
 
 ;;; Code:
 
+(require 'compat)
 (eval-when-compile (require 'cl-lib))
 
 (defgroup orderless nil
@@ -211,6 +212,18 @@ is determined by the values of `completion-ignore-case',
 `read-buffer-completion-ignore-case', as usual for completion."
   :type 'boolean)
 
+(defcustom orderless-expand-substring 'prefix
+  "Whether to perform literal substring expansion.
+This configuration option affects the behavior of some completion
+interfaces when pressing TAB.  If enabled `orderless-try-completion'
+will first attempt literal substring expansion.  If disabled,
+expansion is only performed for single unique matches.  For
+performance reasons only `prefix' expansion is enabled by default.
+Set the variable to `substring' for full substring expansion."
+  :type '(choice (const :tag "No expansion" nil)
+                 (const :tag "Substring" substring)
+                 (const :tag "Prefix (efficient)" prefix)))
+
 ;;; Matching styles
 
 (defun orderless-regexp (component)
@@ -285,7 +298,9 @@ which can invert any predicate or regexp."
   "Return t if STR matches PRED and REGEXP."
   (and str
        (or (not pred) (funcall pred str))
-       (or (not regexp) (string-match-p regexp str))))
+       (or (not regexp)
+           (let ((case-fold-search completion-ignore-case))
+             (string-match-p regexp str)))))
 
 (defun orderless-not (pred regexp)
   "Match strings that do *not* match PRED and REGEXP."
@@ -304,18 +319,13 @@ which can invert any predicate or regexp."
 
 (defun orderless-annotation (pred regexp)
   "Match candidates where the annotation matches PRED and REGEXP."
-  (when-let ((metadata (orderless--metadata))
-             (fun (or (completion-metadata-get
-                       metadata 'annotation-function)
-                      (plist-get completion-extra-properties
-                                 :annotation-function)
-                      (when-let ((aff (or (completion-metadata-get
-                                           metadata 'affixation-function)
-                                          (plist-get completion-extra-properties
-                                                     :affixation-function))))
-                        (lambda (cand) (caddr (funcall aff (list cand))))))))
-    (lambda (str)
-      (orderless--match-p pred regexp (funcall fun str)))))
+  (let ((md (orderless--metadata)))
+    (if-let ((fun (compat-call completion-metadata-get md 'affixation-function)))
+        (lambda (str)
+          (cl-loop for s in (cdar (funcall fun (list str)))
+                   thereis (orderless--match-p pred regexp s)))
+      (when-let ((fun (compat-call completion-metadata-get md 'annotation-function)))
+          (lambda (str) (orderless--match-p pred regexp (funcall fun str)))))))
 
 ;;; Highlighting matches
 
@@ -359,7 +369,6 @@ converted to a list of regexps according to the value of
                   string 'fixedcase 'literal)
                  " +" t)))
 
-(define-obsolete-function-alias 'orderless-dispatch 'orderless--dispatch "1.0")
 (defun orderless--dispatch (dispatchers default string index total)
   "Run DISPATCHERS to compute matching styles for STRING.
 
@@ -456,12 +465,6 @@ string as argument."
    when pred do (cl-callf orderless--predicate-and predicate pred)
    finally return (cons predicate regexps)))
 
-(defun orderless-pattern-compiler (pattern &optional styles dispatchers)
-  "Obsolete function, use `orderless-compile' instead.
-See `orderless-compile' for the arguments PATTERN, STYLES and DISPATCHERS."
-  (cdr (orderless-compile pattern styles dispatchers)))
-(make-obsolete 'orderless-pattern-compiler 'orderless-compile "1.0")
-
 ;;; Completion style implementation
 
 (defun orderless--predicate-normalized-and (p q)
@@ -544,11 +547,10 @@ The predicate PRED is used to constrain the entries in TABLE."
 The predicate PRED is used to constrain the entries in TABLE.  The
 matching portions of each candidate are highlighted.
 This function is part of the `orderless' completion style."
-  (defvar completion-lazy-hilit-fn)
   (pcase-let ((`(,prefix ,regexps ,ignore-case ,pred)
                (orderless--compile string table pred)))
     (when-let ((completions (orderless--filter prefix regexps ignore-case table pred)))
-      (if (bound-and-true-p completion-lazy-hilit)
+      (if completion-lazy-hilit
           (setq completion-lazy-hilit-fn
                 (apply-partially #'orderless--highlight regexps ignore-case))
         (cl-loop for str in-ref completions do
@@ -564,38 +566,43 @@ match, it completes to that match.  If there are no matches, it
 returns nil.  In any other case it \"completes\" STRING to
 itself, without moving POINT.
 This function is part of the `orderless' completion style."
-  (catch 'orderless--many
-    (pcase-let ((`(,prefix ,regexps ,ignore-case ,pred)
-                 (orderless--compile string table pred))
-                (one nil))
-      ;; Abuse all-completions/orderless--filter as a fast search loop.
-      ;; Should be almost allocation-free since our "predicate" is not
-      ;; called more than two times.
-      (orderless--filter
-       prefix regexps ignore-case table
-       (orderless--predicate-normalized-and
-        pred
-        (lambda (arg)
-          ;; Check if there is more than a single match (= many).
-          (when (and one (not (equal one arg)))
-            (throw 'orderless--many (cons string point)))
-          (setq one arg)
-          t)))
-      (when one
-        ;; Prepend prefix if the candidate does not already have the same
-        ;; prefix.  This workaround is needed since the predicate may either
-        ;; receive an unprefixed or a prefixed candidate as argument.  Most
-        ;; completion tables consistently call the predicate with unprefixed
-        ;; candidates, for example `completion-file-name-table'.  In contrast,
-        ;; `completion-table-with-context' calls the predicate with prefixed
-        ;; candidates.  This could be an unintended bug or oversight in
-        ;; `completion-table-with-context'.
-        (unless (or (equal prefix "")
-                    (and (string-prefix-p prefix one)
-                         (test-completion one table pred)))
-          (setq one (concat prefix one)))
-        (or (equal string one) ;; Return t for unique exact match
-            (cons one (length one)))))))
+  (or
+   (pcase orderless-expand-substring
+     ('nil nil)
+     ('prefix (completion-emacs21-try-completion string table pred point))
+     (_ (completion-substring-try-completion string table pred point)))
+   (catch 'orderless--many
+     (pcase-let ((`(,prefix ,regexps ,ignore-case ,pred)
+                  (orderless--compile string table pred))
+                 (one nil))
+       ;; Abuse all-completions/orderless--filter as a fast search loop.
+       ;; Should be almost allocation-free since our "predicate" is not
+       ;; called more than two times.
+       (orderless--filter
+        prefix regexps ignore-case table
+        (orderless--predicate-normalized-and
+         pred
+         (lambda (arg)
+           ;; Check if there is more than a single match (= many).
+           (when (and one (not (equal one arg)))
+             (throw 'orderless--many (cons string point)))
+           (setq one arg)
+           t)))
+       (when one
+         ;; Prepend prefix if the candidate does not already have the same
+         ;; prefix.  This workaround is needed since the predicate may either
+         ;; receive an unprefixed or a prefixed candidate as argument.  Most
+         ;; completion tables consistently call the predicate with unprefixed
+         ;; candidates, for example `completion-file-name-table'.  In contrast,
+         ;; `completion-table-with-context' calls the predicate with prefixed
+         ;; candidates.  This could be an unintended bug or oversight in
+         ;; `completion-table-with-context'.
+         (unless (or (equal prefix "")
+                     (and (string-prefix-p prefix one)
+                          (test-completion one table pred)))
+           (setq one (concat prefix one)))
+         (or (equal string one) ;; Return t for unique exact match
+             (cons one (length one))))))))
 
 ;;;###autoload
 (add-to-list 'completion-styles-alist
